"""
A module for asymmetric's callback schemas.
"""


CALLBACK_OBJECT_SCHEMA = {
    "callback_url_header": {
        "required": False,
        "type": str,
    },
    "callback_method_header": {
        "required": False,
        "type": str,
    },
    "custom_callback_key_header": {
        "required": False,
        "type": str,
    },
}


CALLBACK_OBJECT_DEFAULTS = {
    "callback_url_header": "asymmetric_callback_url",
    "callback_method_header": "asymmetric_callback_method",
    "custom_callback_key_header": "asymmetric_custom_callback_key",
}



"""
A module for containing the loggers of asymmetric.
"""

import json
import logging
import logging.config
import os
from typing import Any, Callable, Dict

from starlette.requests import Request

from asymmetric.constants import LOG_FILE_NAME
from asymmetric.utils import get_body

# Logging configuration
logging.config.dictConfig(
    {
        "version": 1,
        "formatters": {
            "console": {
                "format": "[%(asctime)s] [%(levelname)s] %(module)s: %(message)s"
            },
            "file": {
                "format": (
                    "[%(asctime)s] [%(levelname)s] %(pathname)s - "
                    "line %(lineno)d: \n%(message)s\n"
                )
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
                "formatter": "console",
            },
            "file": {
                "class": "logging.FileHandler",
                "filename": os.getenv("LOG_FILE", default=LOG_FILE_NAME),
                "formatter": "file",
            },
        },
        "root": {"level": "INFO", "handlers": ["console", "file"]},
    }
)


def log(message: str, level: str = "info") -> None:
    """
    Logs {message} with {level} level. Starts the log with '[[asymmetric]]'.
    """
    logger = getattr(logging, level)
    logger(f"[[asymmetric]] {message}")


async def log_request(
    request: Request, route: str, function: Callable[..., Any]
) -> None:
    """
    Logs a request, including the method used for the request, the
    route and the name of the python function being called. Then,
    logs the request body.
    """
    log(
        f"{request.method} request to '{route}' endpoint "
        f"('{function.__name__}' function)."
    )
    log_request_body(await get_body(request))


def log_request_body(body: Dict[str, Any]) -> None:
    """
    Logs a request body formatted as a json.
    """
    log(
        "Request Body:\n"
        + json.dumps(body, indent=2, sort_keys=False, ensure_ascii=False)
    )



"""
A module for asymmetric's core callback logic.
"""

import asyncio
from typing import Any, Callable, Dict, Optional, Union

import httpx
from starlette.datastructures import Headers
from starlette.responses import JSONResponse

from asymmetric.callbacks.callback_object import CALLBACK_OBJECT_DEFAULTS
from asymmetric.callbacks.utils import valid_callback_data
from asymmetric.constants import HTTP_METHODS
from asymmetric.errors import InvalidCallbackHeadersError, InvalidCallbackObjectError
from asymmetric.loggers import log
from asymmetric.utils import generic_call, terminate_program


class CallbackClient:

    """
    A class to abstract the callback logic.
    """

    def __init__(
        self, function: Callable[..., Any], callback: Union[Dict[str, Any], bool]
    ) -> None:
        self.__function = function
        self.__callback = callback
        self.__attribute_finders: Dict[str, str] = {}
        self.__headers = Headers()
        self.__params: Dict[str, str] = {}
        self.__invalid_callback_object = ""

    @property
    def url(self) -> str:
        """Returns the callback URL if it was included, or None."""
        location = self.__attribute_finders.get("callback_url_header")
        return self.__headers.get(location, "") if location is not None else ""

    @property
    def http_method(self) -> str:
        """Returns the callback HTTP method if it was included, or POST."""
        location = self.__attribute_finders.get("callback_method_header", "")
        # POST is the default HTTP method
        http_method = self.__headers[location] if location in self.__headers else "POST"
        return http_method.upper()

    @property
    def custom_key(self) -> Optional[str]:
        """Returns the custom callback key container if it was included, or None."""
        location = self.__attribute_finders.get("custom_callback_key_header", "")
        if location in self.__headers:
            return self.__headers.get(location)
        return None

    def handle_callback(self, headers: Headers, params: Dict[str, Any]) -> JSONResponse:
        """
        Validates that the callback data from the request is correct
        and delegates the main function call. Returns a JSON response.
        """
        if self.__invalid_callback_object:
            # Callback object was defective, server should have stopped
            return JSONResponse(
                {"message": self.__invalid_callback_object}, status_code=500
            )

        try:
            # Set headers and params
            self.__headers = headers
            self.__params = params

            # Validate callback data
            self.__validate_callback_url()
            self.__validate_callback_http_method()

            # Delegate function
            asyncio.ensure_future(self.__callback_call())

            return JSONResponse({}, status_code=202)
        except InvalidCallbackHeadersError as error:
            return JSONResponse({"message": str(error)}, status_code=422)

    def prepare_and_validate_finders(self) -> None:
        """
        Collects the callback data finders and validates
        that they are correct on decoration-time.
        """
        try:
            self.__validate_callback_json_data()
            self.__get_header_finders()
        except InvalidCallbackObjectError as error:
            self.__invalid_callback_object = str(error)
            log(str(error), level="warn")
            terminate_program()

    def __validate_callback_json_data(self) -> None:
        """
        Validates that the callback dictionary is a valid callback
        dictionary or raises an error.
        """
        if not valid_callback_data(self.__callback):
            raise InvalidCallbackObjectError("Invalid callback object")

    def __get_header_finders(self) -> None:
        """
        Gets the header attribute finders to search for some data
        on every request header on decoration-time.
        """
        if isinstance(self.__callback, bool):
            # Callback attribute is True
            self.__attribute_finders = CALLBACK_OBJECT_DEFAULTS
        else:
            # Callback attribute is an object, get callback attribute finders
            for attr, default_value in CALLBACK_OBJECT_DEFAULTS.items():
                if attr in self.__callback:
                    self.__attribute_finders[attr] = self.__callback[attr]
                else:
                    self.__attribute_finders[attr] = default_value

    def __validate_callback_url(self) -> None:
        """
        Validates that the callback URL within the request
        header is valid.
        """
        if not self.url:
            raise InvalidCallbackHeadersError("Invalid callback URL")

    def __validate_callback_http_method(self) -> None:
        """
        Validates that the callback HTTP method within the request
        header is valid.
        """
        if self.http_method.lower() not in HTTP_METHODS:
            raise InvalidCallbackHeadersError("Invalid callback HTTP method")

    async def __callback_call(self) -> None:
        """
        Executes the function and makes the request to the callback endpoint.
        """
        try:
            response = await generic_call(self.__function, self.__params)
            if self.custom_key is not None:
                response = {self.custom_key: response}

            async with httpx.AsyncClient() as client:
                await client.request(
                    self.http_method,
                    self.url,
                    json=response,
                )
        except Exception as error:
            message = "Error while executing the delegated method: "
            message += str(error)
            log(message, level="warn")



"""
Module to hold every custom exception.
"""


class DuplicatedEndpointError(Exception):
    """
    Exception for when an endpoint is already in the added endpoints.
    """


class InvalidCallbackObjectError(Exception):
    """
    Exception for when the callback object is incorrectly formed.
    """


class InvalidCallbackHeadersError(Exception):
    """
    Exception for when the callback headers are incorrect.
    """



"""
A module for every utility of asymmetric.
"""

import inspect
import json
import sys
from typing import Any, Callable, Dict

from starlette.requests import Request
from starlette.responses import JSONResponse


async def generic_call(function: Callable[..., Any], params: Dict[str, Any]) -> Any:
    """
    Executes a function with its params, checking if said function
    is async or not
    """
    if inspect.iscoroutinefunction(function):  # Await async functions
        return await function(**params)
    return function(**params)


def handle_error(error: Exception) -> JSONResponse:
    """Handles errors from the router."""
    return JSONResponse({"message": str(error)}, status_code=500)


def filter_params(function: Callable[..., Any], data: Dict[str, Any]) -> Dict[str, Any]:
    """Filters parameters so that the function recieves only what it needs."""
    # Get the parameters
    params = inspect.getfullargspec(function)
    if params.varkw is not None:
        # The function recieves kwargs, return the full dictionary
        return data
    if not params.args:
        # The function does not recieve args, return an empty dict
        return {}
    # Filter every param whose key is not in the params dictionary
    return {k: v for k, v in data.items() if k in params.args}


async def get_body(request: Request) -> Dict[str, Any]:
    """
    Gets the body of the request and returns an empty dict if the request
    has no body.
    """
    try:
        body = await request.json()
        return body
    except json.decoder.JSONDecodeError:
        return {}


def valid_plain_dict(data: Dict[str, Any], validator: Dict[str, Any]) -> bool:
    """
    Given a data and a validator array, checks if data includes the required
    attributes and if it includes extra attributes.
    """
    possible_attrs = validator.keys()
    required_attrs = filter(lambda k: validator[k]["required"], possible_attrs)
    attrs = data.keys()

    # Check for extra attributes
    if not all(map(lambda x: x in possible_attrs, attrs)):
        return False

    # Check for required attributes
    if not all(map(lambda x: x in attrs, required_attrs)):
        return False

    # Check for invalid types
    if not all(map(lambda k: isinstance(data[k], validator[k]["type"]), attrs)):
        return False

    return True


def terminate_program() -> None:
    """Terminates the server process."""
    sys.exit(1)



"""
The main module of asymmetric.
"""

import asyncio
from typing import Any, Callable, Dict, List, Union

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.types import Receive, Scope, Send

from asymmetric.callbacks.core import CallbackClient
from asymmetric.constants import HTTP_METHODS
from asymmetric.endpoints import Endpoints
from asymmetric.errors import DuplicatedEndpointError
from asymmetric.helpers import http_verb
from asymmetric.loggers import log, log_request
from asymmetric.utils import (
    filter_params,
    generic_call,
    get_body,
    handle_error,
    terminate_program,
)


class Asymmetric:

    """
    Main class to encapsulate every important feature of
    the asymmetric package.
    """

    def __init__(self) -> None:
        self.__app: Starlette = Starlette()
        self.__endpoints: Endpoints = Endpoints()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.__app.__call__(scope, receive, send)

    def __getattr__(self, attr: Any) -> Any:
        """
        Intercept all attribute/method calls to the assymetric object
        if they aren't part of it and redirect them to the Starlette app.
        """
        return getattr(self.__app, attr)

    def router(
        self,
        route: str,
        methods: List[str] = ["post"],
        response_code: int = 200,
        callback: Union[Dict[str, Any], bool] = False,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """
        Method to use for decorating the function wanting to be transformed
        to an API.
        """
        methods = [http_verb(x) for x in methods if http_verb(x) in HTTP_METHODS]

        def decorator(function: Callable[..., Any]) -> Callable[..., Any]:
            """
            Function decorator. Receives the main function and wraps it as a
            starlette endpoint. Returns the original unwrapped function.
            """

            if callback:
                callback_client = CallbackClient(function, callback)
                callback_client.prepare_and_validate_finders()

            @self.__app.route(route, methods=methods)
            async def wrapper(request: Request) -> JSONResponse:
                asyncio.ensure_future(log_request(request, route, function))

                try:
                    # Get the body
                    body = await get_body(request)

                    # Get params and headers
                    params = filter_params(function, body)
                    headers = request.headers

                    if not callback:
                        # Process and return the result
                        return JSONResponse(
                            await generic_call(function, params),
                            status_code=response_code,
                        )

                    return callback_client.handle_callback(headers, params)
                except Exception as error:
                    return handle_error(error)

            # Save Endpoint
            try:
                self.__endpoints.add_endpoints(
                    route,
                    methods,
                    response_code,
                    function,  # Save unchanged function
                    wrapper,  # Save starlette decorated function
                )
            except DuplicatedEndpointError as err:
                log(f"DuplicatedRouteError: {err}", level="error")
                terminate_program()  # TODO: exit the server correctly

            return function

        return decorator


asymmetric_object = Asymmetric()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asymmetric', 'asymmetric.callbacks', 'asymmetric.cli']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.16.1,<0.17.0', 'starlette>=0.13.8,<0.14.0']

entry_points = \
{'console_scripts': ['asymmetric = asymmetric.cli.core:dispatcher']}

setup_kwargs = {
    'name': 'asymmetric',
    'version': '0.1.6',
    'description': 'The async framework that calls you back! Enable ridiculously fast and easy module-to-API transformations. Learn in minutes, implement in seconds.',
    'long_description': '# Asymmetric\n\n![PyPI - Version](https://img.shields.io/pypi/v/asymmetric?style=for-the-badge&logo=python&color=306998&logoColor=%23fff&label=version)\n\n_The async framework that calls you back_! ✨ Enable ridiculously fast and easy module-to-**[API](https://en.wikipedia.org/wiki/Web_API)** transformations. Learn in minutes, implement in seconds.\n\n![Tests Coverage](https://img.shields.io/codecov/c/gh/daleal/asymmetric?color=%2344cc11&label=coverage&logo=codecov&logoColor=ffffff&style=for-the-badge)\n![Linters](https://img.shields.io/github/workflow/status/daleal/asymmetric/linters?label=linters&logo=github&style=for-the-badge)\n\n## Why Asymmetric?\n\nRaw developing speed and ease of use, that\'s why. `asymmetric` is based on **[Starlette](https://github.com/encode/starlette)** ✨! While `Starlette` is a powerful tool to have, getting it to work from scratch can be a bit of a pain, especially if you have never used it before. The idea behind `asymmetric` is to be able to take any module **already written** and transform it into a working API in a matter of minutes, instead of having to design the module ground-up to work with `Starlette` (it can also be used to build an API from scratch really fast). With `asymmetric`, you will also get some neat features, namely:\n\n- Auto logging.\n- Server-side error detection and exception handling.\n- **Asynchronous callback endpoints** to make a request, terminate the request **immediately** and then have the server make a request to a _callback_ endpoint with the results! ✨\n- ~~Auto-generated `/docs` endpoint for your API with **interactive documentation**.~~ **[UNDER CONSTRUCTION]**\n- ~~Auto-generated [OpenAPI Specification](https://swagger.io/docs/specification/about/) documentation files for your API.~~ **[UNDER CONSTRUCTION]**\n\nThe [complete documentation](https://asymmetric.one/docs/) is available on the [official website](https://asymmetric.one/).\n\n## Installing\n\nInstall using pip!\n\n```sh\npip install asymmetric\n```\n\n## Usage\n\n### Running the development server\n\nTo start a server, choose your favorite `ASGI` server and target the `asymmetric` object!\n\n```sh\nuvicorn <module>:asymmetric\n```\n\nWhere `<module>` is your module name (in the examples, we will be writing in a file named `module.py`, so the module name will be just `module`). A `Starlette` instance will be spawned immediately and can be reached at [http://127.0.0.1:8000](http://127.0.0.1:8000) by default. We don\'t have any endpoints yet, so we\'ll add some later.\n\n### Defining the API endpoints\n\nThe module consists of a main object called `asymmetric`, which includes an important element: the `router` decorator. Let\'s analyze it:\n\n```py\nfrom asymmetric import asymmetric\n\n@asymmetric.router("/some-route", methods=["post"], response_code=200, callback=False)\n```\n\nThe decorator recieves 4 arguments: the `route` argument (the endpoint of the API to which the decorated function will map), the `methods` argument (a list of the methods accepted to connect to that endpoint, defaults in only `POST` requests), the `response_code` argument (the response code of the endpoint if everything goes according to the plan. Defaults to `200`) and the `callback` argument (a boolean or an object specifying the request style for that endpoint, defaults to `False`, generating normal endpoint behaviour). The `callback` attribute will have its own section below, for now we will use the default `callback=False`.\n\nNow let\'s imagine that we have the following method:\n\n```py\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n```\n\nTo transform that method into an API endpoint, all you need to do is add one line:\n\n```py\n@asymmetric.router("/sample", methods=["get"])\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n```\n\nRun `uvicorn module:asymmetric` and send a `GET` request to `http://127.0.0.1:8000/sample`. You should get a `Hello World!` in response! (To try it with a browser, make sure to run the above command and click [this link](http://127.0.0.1:8000/sample)).\n\nBut what about methods with arguments? Of course they can be API\'d too! Let\'s now say that you have the following function:\n\n```py\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n```\n\nTo transform that method into an API endpoint, all you need to do, again, is add one line:\n\n```py\n@asymmetric.router("/add")\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n```\n\n### Querying API endpoints\n\nTo give parameters to a function, all we need to do is send a `json` body with the names of the parameters as keys. Let\'s see how! Run `uvicorn module:asymmetric` and send a `POST` request (the default `HTTP` method) to `http://127.0.0.1:8000/add`, now using the `httpx` module.\n\n```python\nimport httpx\n\npayload = {\n    "a": 48,\n    "b": 21\n}\nresponse = httpx.post("http://127.0.0.1:8000/add", json=payload)\nprint(response.json())\n```\n\nWe got a `69` response! (`48 + 21 = 69`). Of course, you can return dictionaries from your methods and those will get returned as a `json` body in the response object **automagically**!\n\nWith this in mind, you can transform any existing project into a usable API very quickly!\n\n## What about `async`?\n\nGiven that the underlying framework is `Starlette`, you can use `async` to define your methods, no problem! Here\'s an example:\n\n```py\n@asymmetric.router("/another-add")\nasync def another_async_function(a, b=372):\n    """\n    Adds :a and :b asynchronously and returns the\n    result of that operation.\n    """\n    return a + b\n```\n\n## Call me back!\n\nDon\'t you hate it when people don\'t call you back after a date? We all have lived that annoying experience. But don\'t worry! `asymmetric` **will** call you back!\n\nSome functions may be **too heavy** to be executed to respond to an `HTTP` request. Maybe your function is a predictor of some sort, and it requires an hour of processing time to spit out results. Here\'s when the `callback` parameter of the `asymmetric` decorator comes into play! You can ask `asymmetric` to terminate the `HTTP` request **immediately**, keep processing stuff and then, once it finishes, **execute a request to a specified endpoint with the results**. Let\'s imagine that we have a `predict` endpoint that we want to transform into an `API`:\n\n```python\ndef predict(data):\n    values = Model.predict(data)\n\n    # One hour later...\n    return values\n```\n\nJust add the `asymmetric` decorator and you\'re good to go!\n\n```python\n@asymmetric.router("/predict", callback=True)\ndef predict(data):\n    values = Model.predict(data)\n\n    # One hour later...\n    return values\n```\n\nOf course, if you rely on some `async` sorcery for your operations, `asymmetric` can handle it!\n\n```python\n@asymmetric.router("/predict", callback=True)\nasync def predict(data):\n    values = await Model.predict(data)\n\n    # One hour later...\n    return values\n```\n\nStart the server with `uvicorn module:asymmetric` and now you are able to call the endpoint using the following snippet:\n\n```py\nimport httpx\n\nresponse = httpx.post(\n    "http://127.0.0.1:8000/predict",\n    json={"data": mydata},\n    headers={\n        "asymmetric_callback_url": "http://callback.url/receive/predictions",\n        "asymmetric_callback_method": "post",\n    }\n)\n\nprint(response)\n```\n\nWow... **What?!** You just witnessed **the magic of `asymmetric`**. The response will be available **immediately** with a `202` status code. Meanwhile, the server will keep processing the request. When it finishes, **it will make a `POST` request to the endpoint specified in the headers** with the content of the method\'s return value. Cool, right? But what if I want to send the content of the method\'s return value inside a `json`, as the value of a `predictions` key? Well, that\'s easy! Just change the headers!\n\n```py\nimport httpx\n\nresponse = httpx.post(\n    "http://127.0.0.1:8000/predict",\n    json={"data": mydata},\n    headers={\n        "asymmetric_callback_url": "http://callback.url/receive/predictions",\n        "asymmetric_callback_method": "post",\n        "asymmetric_custom_callback_key": "predictions",\n    }\n)\n\nprint(response)\n```\n\nThat will send a `json` with one element, with `predictions` as a key and the result of the function as the value. **The key here are the headers**. They specify what to do with the result of your function. **You can also change the required headers, if you want to!**\n\n```python\ncallback_parameters = {\n    "callback_url_header": "send_me_here",\n    "callback_method_header": "use_me",\n    "custom_callback_key_header": "put_me_in_here",\n}\n\n@asymmetric.router("/predict", callback=callback_parameters)\nasync def predict(data):\n    values = await Model.predict(data)\n\n    # One hour later...\n    return values\n```\n\nNow, to achieve the same result as before, the requests must change their headers!\n\n```py\nimport httpx\n\nresponse = httpx.post(\n    "http://127.0.0.1:8000/predict",\n    json={"data": mydata},\n    headers={\n        "send_me_here": "http://callback.url/receive/predictions",\n        "use_me": "post",\n        "put_me_in_here": "predictions",\n    }\n)\n\nprint(response)\n```\n\nAs you probably imagine by now, the `callback` parameter can be a boolean or a dictionary with the following schema:\n\n```json\n{\n    "type": "object",\n    "properties": {\n        "callback_url_header": {\n            "type": "string"\n        },\n        "callback_method_header": {\n            "type": "string"\n        },\n        "custom_callback_key_header": {\n            "type": "string"\n        }\n    }\n}\n```\n\nIf no `HTTP` method is specified, the server will `POST` the information to the callback `URL`.\n\n## ~~ReDoc Documentation~~\n\n**[UNDER CONSTRUCTION]**\n\nBy default, you can `GET` the `/docs` endpoint (using a browser) to access to **interactive auto-generated documentation** about your API. It will include request bodies for each endpoint, response codes, authentication required, default values, and much more!\n\n**Tip**: Given that the [ReDoc Documentation](https://github.com/Redocly/redoc) is based on the OpenAPI standard, using **type annotations** in your code will result in a more detailed interactive documentation. Instead of the parameters being allowed to be any type, they will be forced into the type declared in your code. Cool, right?\n\n## To Do\n\n- _Automagic_ OpenAPI spec isn\'t being generated rigth now, so that\'s missing from the library. It will soon be added, though, as it is a very useful _feature_.\n- Parse callback `URL`s to make sure that they are valid `URL`s, and fail if they aren\'t.\n- On some initialization errors, the server should stop to _avoid avoidable errors_. Right now, the method to stop the server really does nothing, so that\'s something that should be addressed in the near future.\n\n## Developing\n\nClone the repository:\n\n```sh\ngit clone https://github.com/daleal/asymmetric.git\n\ncd asymmetric\n```\n\nRecreate environment:\n\n```sh\nmake get-poetry\nmake venv-with-dependencies\n```\n\n## Resources\n\n- [Official Website](https://asymmetric.one/)\n- [Issue Tracker](https://github.com/daleal/asymmetric/issues/)\n',
    'author': 'Daniel Leal',
    'author_email': 'dlleal@uc.cl',
    'maintainer': 'Daniel Leal',
    'maintainer_email': 'dlleal@uc.cl',
    'url': 'https://asymmetric.one/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
A module to route the CLI traffic.
"""

from argparse import ArgumentParser

import asymmetric


def dispatcher() -> None:
    """
    Main CLI method, recieves the command line action and dispatches it to
    the corresponding method.
    """
    parser = generate_parser()
    parser.parse_args()


def generate_parser() -> ArgumentParser:
    """Generates the CLI parser for the module."""
    # Create parser
    parser = ArgumentParser(description="Command line interface tool for asymmetric.")

    # Add version command
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"asymmetric version {asymmetric.__version__}",
    )

    return parser



"""
A module for every helper of asymmetric.
"""


def http_verb(dirty: str) -> str:
    """
    Given a 'dirty' HTTP verb (uppercased, with trailing whitespaces), strips
    it and returns it lowercased.
    """
    return dirty.strip().lower()
